@@ -124,7 +124,7 @@ http_verify(const char *header, const char *cip, const int sfd)
 		if (c[sfd].x_ssl) {
 			SSL_write(c[sfd].ssl, b, strlen(b));
 		} else {
-			send(sfd, b, strlen(b), 0);
+			sys_write(sfd, b, strlen(b));
 		}
 		free(h);
 		free(b);
@@ -140,14 +140,15 @@ http_verify(const char *header, const char *cip, const int sfd)
  *	0 = close connection, 1 = keep alive connection, -1 = error
  */
 int
-http_proc(const char *header, char *body, int hr, int sfd)
+http_proc(const char *header, char *body, const int hr, const int sfd)
 {
 	DIR		*odir;
 	int		cpid, file, fds1[2], fds2[2];
-	int		i, r, s, bs, rd, sp, st, len, frk;
-	char		*x, *h = NULL;
+	int		i, r, s, bs, rd, sp, st, len, size;
+	us_int		x_nph, x_fork, x_cpage;
+	char		*h = NULL, *x = NULL;
 	char		ch[1], image[128], tmp[1024], full[1024], status[1024];
-	char		b[BS];
+	char		b[BS], cpage[1024];
 	struct tm	*t = NULL;
 	struct stat	sta;
 	struct dirent	*rdir;
@@ -155,7 +156,7 @@ http_proc(const char *header, char *body, int hr, int sfd)
 	struct timeval	tv;
 	fd_set		o_read;
 
-	r = sp = st = frk = 0;
+	r = sp = st = x_nph = x_fork = x_cpage = 0;
 	bs = BS;
 
 	/* parse request header */
@@ -170,7 +171,7 @@ http_proc(const char *header, char *body, int hr, int sfd)
 		if (c[sfd].x_ssl) {
 			SSL_write(c[sfd].ssl, x, strlen(x));
 		} else {
-			send(sfd, x, strlen(x), 0);
+			sys_write(sfd, x, strlen(x));
 		}
 		free(x);
 	}
@@ -179,35 +180,116 @@ http_proc(const char *header, char *body, int hr, int sfd)
 			SSL_write(c[sfd].ssl, rh->rp_header,
 			    strlen(rh->rp_header));
 		} else {
-			send(sfd, rh->rp_header, strlen(rh->rp_header), 0);
+			sys_write(sfd, rh->rp_header, strlen(rh->rp_header));
 		}
 	}
 	if (rh->rp_status == 401) {
 		x = http_body(http_s_401, "", rh->rp_header);
-		if (c[sfd].x_ssl) {
-			SSL_write(c[sfd].ssl, x, strlen(x));
+		if (config.c401[0] != '0') {
+			if (rh->x_dov) {
+				snprintf(cpage, sizeof(cpage), "%s/%s",
+				    rh->rq_docrootv, config.c401);
+			} else {
+				snprintf(cpage, sizeof(cpage), "%s/%s",
+				    config.docroot, config.c401);
+			}
+			x_cpage = 1;
 		} else {
-			send(sfd, x, strlen(x), 0);
+			if (c[sfd].x_ssl) {
+				SSL_write(c[sfd].ssl, x, strlen(x));
+			} else {
+				sys_write(sfd, x, strlen(x));
+			}
+			free(x);
 		}
-		free(x);
 	}
 	if (rh->rp_status == 403) {
 		x = http_body(http_s_403, "", rh->rp_header);
-		if (c[sfd].x_ssl) {
-			SSL_write(c[sfd].ssl, x, strlen(x));
+		if (config.c403[0] != '0') {
+			if (rh->x_dov) {
+				snprintf(cpage, sizeof(cpage), "%s/%s",
+				    rh->rq_docrootv, config.c403);
+			} else {
+				snprintf(cpage, sizeof(cpage), "%s/%s",
+				    config.docroot, config.c403);
+			}
+			x_cpage = 1;
 		} else {
-			send(sfd, x, strlen(x), 0);
+			if (c[sfd].x_ssl) {
+				SSL_write(c[sfd].ssl, x, strlen(x));
+			} else {
+				sys_write(sfd, x, strlen(x));
+			}
+			free(x);
 		}
-		free(x);
 	}
 	if (rh->rp_status == 404) {
 		x = http_body(http_s_404, "", rh->rp_header);
-		if (c[sfd].x_ssl) {
-			SSL_write(c[sfd].ssl, x, strlen(x));
+		if (config.c404[0] != '0') {
+			if (rh->x_dov) {
+				snprintf(cpage, sizeof(cpage), "%s/%s",
+				    rh->rq_docrootv, config.c404);
+			} else {
+				snprintf(cpage, sizeof(cpage), "%s/%s",
+				    config.docroot, config.c404);
+			}
+			x_cpage = 1;
+		} else  {
+			if (c[sfd].x_ssl) {
+				SSL_write(c[sfd].ssl, x, strlen(x));
+			} else {
+				sys_write(sfd, x, strlen(x));
+			}
+			free(x);
+		}
+	}
+
+	/*
+	 * Custom response
+	 */
+	if (x_cpage) {
+		/* open custom file */
+		if ((file = open(cpage, O_RDONLY, 0)) == -1) {
+			/* can not open custom file, send default page */
+			if (c[sfd].x_ssl) {
+				SSL_write(c[sfd].ssl, x, strlen(x));
+			} else {
+				sys_write(sfd, x, strlen(x));
+			}
+			free(x);
 		} else {
-			send(sfd, x, strlen(x), 0);
+			free(x);
+			/* custom file opened, send it */
+			stat(cpage, &sta);
+			memset(b, 0, sizeof(b));
+			read(file, b, bs);
+			x = http_chunk(b);
+
+			size = rh->rp_hsize + strlen(x) + strlen(http_fv_lch)
+			    + 1;
+
+			h = malloc(size);
+			strlcpy(h, rh->rp_header, size);
+			strlcat(h, x, size);
+			if (sta.st_size <= bs) {
+				strlcat(h, http_fv_lch, size);
+			} else {
+				c[sfd].pfdn[hr] = file;
+				c[sfd].x_chk[hr] = 1;
+				c[sfd].pfdo++;
+			}
+
+			if (c[sfd].x_ssl) {
+				SSL_write(c[sfd].ssl, h, strlen(h));
+			} else {
+				sys_write(sfd, h, strlen(h));
+			}
+
+			free(h);
+			free(x);
+			close(file);
+			r = 1;
 		}
-		free(x);
 	}
 
 	/*
@@ -242,7 +324,7 @@ http_proc(const char *header, char *body, int hr, int sfd)
 				return -1;
 			}
 		} else {
-			if ((sp = send(sfd, h, rh->rp_hsize + rd, 0)) == -1) {
+			if ((sp = sys_write(sfd, h, rh->rp_hsize + rd)) == -1) {
 				free(h);
 				free(rh);
 				close(file);
@@ -258,11 +340,11 @@ http_proc(const char *header, char *body, int hr, int sfd)
 			close(file);
 		} else {
 			/* first file block sent, save infos and return */
-			c[sfd].pfdn[hr] = file;
-			c[sfd].pfds[hr] = st;
 			c[sfd].pfdo++;
 			c[sfd].state = rh->x_con;
-			c[sfd].status = rh->rp_status;
+			c[sfd].pfdn[hr] = file;
+			c[sfd].pfds[hr] = st;
+			c[sfd].status[hr] = rh->rp_status;
 			c[sfd].plreq[hr] = malloc(128);
 			c[sfd].plref[hr] = malloc(128);
 			c[sfd].plagt[hr] = malloc(128);
@@ -294,7 +376,7 @@ http_proc(const char *header, char *body, int hr, int sfd)
 	if ((rh->rp_status == 200 || rh->rp_status == 206) && rh->x_cgi == 0 &&
 	    rh->x_idx == 0 && strcmp(rh->rq_method, "HEAD") == 0) {
 		/* send header */
-		send(sfd, rh->rp_header, rh->rp_hsize, 0);
+		sys_write(sfd, rh->rp_header, rh->rp_hsize);
 
 		/* no body has been sent */
 		st = 0;
@@ -315,7 +397,10 @@ http_proc(const char *header, char *body, int hr, int sfd)
 			free(rh);
 			return 0;
 		}
-		frk = 1;
+		x_fork = 1;
+
+		/* close all copied fds from parent */
+		sys_close_except(sfd);
 
 		/* create pipes to communicate with cgi */
 		pipe(fds1);
@@ -379,14 +464,20 @@ http_proc(const char *header, char *body, int hr, int sfd)
 					    strerror(errno));
 					exit(1);
 				}
-				strlcpy(x, rh->rp_header, len);
-				strlcat(x, b, len);
+				/* check if it is a nph cgi */
+				if (strncasecmp("nph-", rh->rq_files, 4) == 0) {
+					strlcpy(x, b, len);
+					x_nph = 1;
+				} else {
+					strlcpy(x, rh->rp_header, len);
+					strlcat(x, b, len);
+				}
 				if (c[sfd].x_ssl) {
 					if ((sp = SSL_write(c[sfd].ssl, x,
 					    strlen(x))) < 1)
 						exit(1);
 				} else {
-					if ((sp = send(sfd, x, strlen(x), 0))
+					if ((sp = sys_write(sfd, x, strlen(x)))
 					    < 1)
 						exit(1);
 				}
@@ -400,7 +491,7 @@ http_proc(const char *header, char *body, int hr, int sfd)
 		FD_SET(fds2[0], &o_read);
 
 		/* cgi loop */
-		for (;;) {
+		while (1) {
 			tv.tv_sec = TOC;
 			tv.tv_usec = 0;
 			if (select(fds2[0] + 1, &o_read, NULL, NULL, &tv) != 1)
@@ -411,26 +502,34 @@ http_proc(const char *header, char *body, int hr, int sfd)
 			memset(b, 0, sizeof(b));
 			if ((rd = read(fds2[0], b, sizeof(b))) < 1)
 				break;
-			x = http_chunk(b);
+			if (x_nph)
+				x = b;
+			else
+				x = http_chunk(b);
 			if (c[sfd].x_ssl) {
 				if ((sp = SSL_write(c[sfd].ssl, x, strlen(x)))
 				    < 1)
 					break;
 			} else {
-				if ((sp = send(sfd, x, strlen(x), 0)) < 1)
+				if ((sp = sys_write(sfd, x, strlen(x))) < 1)
 					break;
 			}
 			st += sp;
-			free(x);
+			if (!x_nph)
+				free(x);
 		}
 
 		/* send last chunk */
-		if (c[sfd].x_ssl) {
-			if ((sp = SSL_write(c[sfd].ssl, "0\r\n\r\n", 5)) != -1)
+		if (!x_nph) {
+			if (c[sfd].x_ssl) {
+				if ((sp = SSL_write(c[sfd].ssl, http_fv_lch,
+			    	    strlen(http_fv_lch))) != -1)
 				st += sp;
-		} else {
-			if ((sp = send(sfd, "0\r\n\r\n", 5, 0)) != -1)
+			} else {
+				if ((sp = sys_write(sfd, http_fv_lch,
+			    	    strlen(http_fv_lch))) != -1)
 				st += sp;
+			}
 		}
 
 		/* close fd to cgis stdout */
@@ -455,14 +554,17 @@ http_proc(const char *header, char *body, int hr, int sfd)
 			free(rh);
 			return 0;
 		}
-		frk = 1;
+		x_fork = 1;
+
+		/* close all copied fds from parent */
+		sys_close_except(sfd);
 
 		/* send header */
 		if (c[sfd].x_ssl) {
 			SSL_write(c[sfd].ssl, rh->rp_header,
 			    strlen(rh->rp_header));
 		} else {
-			send(sfd, rh->rp_header, strlen(rh->rp_header), 0);
+			sys_write(sfd, rh->rp_header, strlen(rh->rp_header));
 		}
 
 		/* open directory */
@@ -515,8 +617,8 @@ http_proc(const char *header, char *body, int hr, int sfd)
 					    strlen(x))) == -1)
 						break;
 				} else {
-					if ((sp = send(sfd, x, strlen(x), 0))
-					    == -1)
+					if ((sp = sys_write(sfd, x, strlen(x)))
+					     == -1)
 						break;
 				}
 				st += sp;
@@ -539,7 +641,7 @@ http_proc(const char *header, char *body, int hr, int sfd)
 					!= 1)
 					st += sp;
 			} else {
-				if ((sp = send(sfd, x, strlen(x), 0)) != 1)
+				if ((sp = sys_write(sfd, x, strlen(x))) != 1)
 					st += sp;
 			}
 			memset(b, 0, sizeof(b));
@@ -554,18 +656,19 @@ http_proc(const char *header, char *body, int hr, int sfd)
 				    != -1)
 					st += sp;
 			} else {
-				if ((sp = send(sfd, x, strlen(x), 0)) != -1)
+				if ((sp = sys_write(sfd, x, strlen(x))) != -1)
 					st += sp;
 			}
 			free(x);
 
 			/* send last chunk */
 			if (c[sfd].x_ssl) {
-				if ((sp = SSL_write(c[sfd].ssl, "0\r\n\r\n", 5))
-				    != -1)
+				if ((sp = SSL_write(c[sfd].ssl, http_fv_lch,
+				    strlen(http_fv_lch))) != -1)
 					st += sp;
 			} else {
-				if ((sp = send(sfd, "0\r\n\r\n", 5, 0)) != -1)
+				if ((sp = sys_write(sfd, http_fv_lch,
+				    strlen(http_fv_lch))) != -1)
 					st += sp;
 			}
 		}
@@ -598,7 +701,7 @@ http_proc(const char *header, char *body, int hr, int sfd)
 	flog(logn, "%s %s\n", log_1, log_2);
 
 	/* nhttpd child exits */
-	if (frk)
+	if (x_fork)
 		exit(0);
 
 	free(rh);
@@ -757,24 +860,32 @@ http_header_comp(const char *header)
 int
 http_body_comp(const char *header, char *body)
 {
-	int	i, r;
+	int	i, r, len;
 	char	line[1024], option[1024], clength[16];
 	
 	r = 0;
 
-	/* some browsers send final '\n' at body end */
-	if (body[strlen(body) - 1] == '\n')
-		body[strlen(body) - 1] = '\0';
+	/* get body size */
+	len = strlen(body);
 
+	/* cut out body termination crap */
+	if (body[len - 2] == '\r') {
+		body[len - 2] = '\0';
+		len -= 2;
+	} else if (body[len - 1] == '\n') {
+		body[len - 1] = '\0';
+		len -= 1;
+	}
+
+	/* get Content-Length: from header */
 	for (i = 2; strcutl(line, header, i, sizeof(line)) != -1; i++) {
 		strcutw(option, line, 1, sizeof(option));
-		/* Content-Length: */
 		if (strcasecmp(option, http_fn_clt) == 0)
 			strcuts(clength, line, ' ', '\0', sizeof(clength));
 	}
 
 	/* some browsers count final '\n' to Content-Length, some not, sigh */
-	if (strlen(body) == atoi(clength) || strlen(body) + 1 == atoi(clength))
+	if (len == atoi(clength) || len + 1 == atoi(clength))
 		r = 1;
 
 	return r;
@@ -824,10 +935,10 @@ http_access_htaccess(char *dst, const char *rootdir, const int dsize)
  * http_alog()
  *	write access log after partitial file send
  * Return:
- *	0 = successfull, -1 = error
+ *	0 = success
  */
 int
-http_alog(int sfd, int nr)
+http_alog(const int sfd, const int hr)
 {
 	int		gmt;
 	time_t		tnow;
@@ -839,12 +950,12 @@ http_alog(int sfd, int nr)
 	t = localtime(&tnow);
 
 	snprintf(log_1, sizeof(log_1), "%s - - %s \"%s\" %d %d",
-	    c[sfd].ip, sys_date(t, gmt), c[sfd].plreq[nr], c[sfd].status,
-	    c[sfd].pfds[nr]);
+	    c[sfd].ip, sys_date(t, gmt), c[sfd].plreq[hr], c[sfd].status[hr],
+	    c[sfd].pfds[hr]);
 	snprintf(log_2, sizeof(log_2), "\"%s\" \"%s\"",
-	    c[sfd].plref[nr], c[sfd].plagt[nr]);
+	    c[sfd].plref[hr], c[sfd].plagt[hr]);
 
-	flog(c[sfd].pllog[nr], "%s %s\n", log_1, log_2);
+	flog(c[sfd].pllog[hr], "%s %s\n", log_1, log_2);
 
 	return 0;
 }
@@ -1032,7 +1143,6 @@ http_body(const char *title, const char *msg, const char *header)
 {
 	int	size;
 	char	*rp, *body_chunk, body[8192];
-	char	*last_chunk = "0\r\n\r\n";
 
 	/* html */
 	snprintf(body, sizeof(body), "<html>\n<head>\n<title>%s</title>\n"
@@ -1043,7 +1153,7 @@ http_body(const char *title, const char *msg, const char *header)
 	body_chunk = http_chunk(body);
 
 	/* allocate memory for the response */
-	size = strlen(header) + strlen(body_chunk) + strlen(last_chunk) + 1;
+	size = strlen(header) + strlen(body_chunk) + strlen(http_fv_lch) + 1;
 	if ((rp = malloc(size)) == NULL)
 		return NULL;
 	memset(rp, 0, size);
@@ -1051,7 +1161,7 @@ http_body(const char *title, const char *msg, const char *header)
 	/* create response */
 	strlcpy(rp, header, size);
 	strlcat(rp, body_chunk, size);
-	strlcat(rp, last_chunk, size);
+	strlcat(rp, http_fv_lch, size);
 
 	free(body_chunk);
 
@@ -1072,7 +1182,7 @@ http_header(const char *header_data, const char *force_status,
 	int		fsize, fd, i, r;
 	char		line[1024], option[1024];
 	char		alias[1024], cgi_path[1024], cgi_full[1024];
-	char		docroot[1024], docrootn[1024], file_path[1024];
+	char		docroot[1024], file_path[1024];
 	char		acc_base64[1024], acc_file[1024];
 	char		acc_userpw[1024], acc_user[1024], acc_pw[1024];
 	char		tmp[1024], status[128];
@@ -1100,6 +1210,7 @@ http_header(const char *header_data, const char *force_status,
 	h->x_ran = 0;
 	h->x_cgi = 0;
 	h->x_idx = 0;
+	h->x_dov = 0;
 	h->rp_status = 200;
 	strlcpy(status, http_s_200, sizeof(status));
 
@@ -1191,17 +1302,18 @@ http_header(const char *header_data, const char *force_status,
 	}
 
 	/* set default docroot and access log */
-	strlcpy(docrootn, config.docroot, sizeof(docrootn));
+	strlcpy(h->rq_docrootv, config.docroot, sizeof(h->rq_docrootv));
 	strlcpy(logn, config.logaccess, sizeof(logn));
 
 	/* virtual hosts */
 	if (h->x_hos == 1 && strcmp(http_url, h->rq_fv_hos) != 0) {
 		/* search for virutal host in configuration */
-		if (fparse(docrootn, h->rq_fv_hos, config.file,
-		    sizeof(docrootn)) > 0) {
+		if (fparse(h->rq_docrootv, h->rq_fv_hos, config.file,
+		    sizeof(h->rq_docrootv)) > 0) {
 			/* set own access_log for virutal host */
 			snprintf(logn, sizeof(logn), "%s_%s", config.logaccess,
 			    h->rq_fv_hos);
+			h->x_dov = 1;
 		}
 	}
 
@@ -1251,12 +1363,12 @@ http_header(const char *header_data, const char *force_status,
 			    strlen(h->rq_uri), sizeof(file_path));
 		} else {
 			/* no keep defaults */
-			strlcpy(docroot, docrootn, sizeof(docroot));
+			strlcpy(docroot, h->rq_docrootv, sizeof(docroot));
 			strlcpy(file_path, h->rq_uri, sizeof(file_path));
 		}
 
 		/* directory? */
-		if (h->rq_uri[strlen(h->rq_uri)-1] == '/') {
+		if (h->rq_uri[strlen(h->rq_uri) - 1] == '/') {
 			/* save original uri */
 			strlcpy(h->rq_index, file_path, sizeof(h->rq_index));
 
