@@ -98,10 +98,11 @@ const char *http_fn_aut		= "Authorization:";
 const char *http_fn_auw		= "WWW-Authenticate:";
 const char *http_fn_ran		= "Range:";
 const char *http_fn_rac		= "Content-Range:";
-const char *http_fv_srv		= "Nostromo/1.7.1 (Unix)";
+const char *http_fv_srv		= "Nostromo/1.7.2 (Unix)";
 const char *http_fv_pro		= "HTTP/1.1";
 const char *http_fv_cgi		= "CGI/1.1";
 const char *http_fv_teg		= "chunked";
+const char *http_fv_lch		= "0\r\n\r\n";
 const char *http_fv_con_alv	= "Keep-Alive";
 const char *http_fv_con_cls	= "close";
 
@@ -187,6 +188,94 @@ sig_handler(const int sig)
 	}
 }
 
+/*
+ * send_chunk()
+ *	send a file in chunks
+ * Return:
+ *	none
+ */
+void
+send_chunk(const int sfd, const int hr)
+{
+	int		r, s;
+	char		*b;
+	static char	fbuf[BS];
+
+	r = s = 0;
+
+	/* we need to initialize buf to do nice chunks */
+	memset(fbuf, 0, sizeof(fbuf));
+
+	/* read block from file */
+	r = read(c[sfd].pfdn[hr], fbuf, BS);
+
+	/* send a chunk */
+	if (r > 0) {
+		b = http_chunk(fbuf);
+		if (c[sfd].x_ssl) {
+			s = SSL_write(c[sfd].ssl, b, strlen(b));
+		} else {
+			s = sys_write(sfd, b, strlen(b));
+		}
+		free(b);
+	}
+
+	/* send last chunk */
+	if (r < 1 || s < 1) {
+		if (c[sfd].x_ssl) {
+			s = SSL_write(c[sfd].ssl, http_fv_lch,
+			    strlen(http_fv_lch));
+		} else {
+			s = sys_write(sfd, http_fv_lch, strlen(http_fv_lch));
+		}
+		close(c[sfd].pfdn[hr]);
+		c[sfd].x_chk[hr] = 0;
+		c[sfd].pfdo--;
+	}
+}
+
+/*
+ * send_file()
+ *	send a file in blocks, the size is delivered in the header
+ * Return:
+ *	none
+ */
+void
+send_file(const int sfd, const int hr)
+{
+	int		r, s;
+	static char	fbuf[BS];
+
+	r = s = 0;
+
+	/* read block from file */
+	r = read(c[sfd].pfdn[hr], fbuf, BS);
+
+	/* send a block */
+	if (r > 0) {
+		if (c[sfd].x_ssl) {
+			s = SSL_write(c[sfd].ssl, fbuf, r);
+		} else {
+			s = sys_write(sfd, fbuf, r);
+		}
+		if (s != -1)
+			c[sfd].pfds[hr] += s;
+	}
+
+	/* file sent, write access log, cleanup */
+	if (r < 1 || s < 1) {
+		http_alog(sfd, hr);
+		close(c[sfd].pfdn[hr]);
+		c[sfd].pfdn[hr] = 0;
+		c[sfd].pfds[hr] = 0;
+		c[sfd].pfdo--;
+		free(c[sfd].plreq[hr]);
+		free(c[sfd].plref[hr]);
+		free(c[sfd].plagt[hr]);
+		free(c[sfd].pllog[hr]);
+	}
+}
+
 /*
  * main()
  *	Nostromo webserver
@@ -201,8 +290,8 @@ main(int argc, char *argv[])
 	int			i, j, k, r, s, ch, rd, rp, rt, wr, post, size;
 	int			fd, erro, mode, opt_c, opt_r, opt_4, opt_6;
 	char			in[HDN * HDS + 1], tmp[HDS + 1], body[HDS + 1];
-	char			header[HDN][HDS + 1], fbuf[BS];
-	char			*h, *b;
+	char			header[HDN][HDS + 1];
+	char			*b, *h;
 	extern char		**environ;
 	struct passwd		*pwd = NULL;
 	struct stat		st;
@@ -362,6 +451,15 @@ main(int argc, char *argv[])
 			r = -1;
 		}
 	}
+	if (fparse(config.c401, "custom_401", config.file, sizeof(config.user))
+	    == -1)
+		strlcpy(config.c401, "0", sizeof(config.c401));
+	if (fparse(config.c403, "custom_403", config.file, sizeof(config.user))
+	    == -1)
+		strlcpy(config.c403, "0", sizeof(config.c403));
+	if (fparse(config.c404, "custom_404", config.file, sizeof(config.user))
+	    == -1)
+		strlcpy(config.c404, "0", sizeof(config.c404));
 	/* we need all mandatory parameters otherwise no start */
 	if (r == -1)
 		exit(1);
@@ -741,7 +839,7 @@ main(int argc, char *argv[])
 	/*
 	 * Main loop
 	 */
-	for (;;) {
+	while (1) {
 		/* copy master set */
 		set_r = master_r;
 		set_w = master_w;
@@ -837,7 +935,7 @@ main(int argc, char *argv[])
 						SSL_free(c[sdnew].ssl);
 						c[sdnew].x_ssl = 0;
 					} else {
-						send(sdnew, b, strlen(b), 0);
+						sys_write(sdnew, b, strlen(b));
 					}
 					free(h);
 					free(b);
@@ -886,30 +984,12 @@ main(int argc, char *argv[])
 				for (i = 0; i < HDN; i++) {
 					if (c[sdnow].pfdn[i] == 0)
 						continue;
-					r = read(c[sdnow].pfdn[i], fbuf, BS);
-					if (r > 0) {
-						if (c[sdnow].x_ssl) {
-							s = SSL_write(
-							    c[sdnow].ssl, fbuf,
-							    r);
-						} else {
-							s = send(sdnow, fbuf, r,
-							    0);
-						}
-						if (s != -1)
-							c[sdnow].pfds[i] += s;
-					}
-					if (r < 1 || s < 1) {
-						http_alog(sdnow, i);
-						close(c[sdnow].pfdn[i]);
-						c[sdnow].pfdn[i] = 0;
-						c[sdnow].pfds[i] = 0;
-						c[sdnow].pfdo--;
-						free(c[sdnow].plreq[i]);
-						free(c[sdnow].plref[i]);
-						free(c[sdnow].plagt[i]);
-						free(c[sdnow].pllog[i]);
-					}
+					/* send data chunked */
+					if (c[sdnow].x_chk[i])
+						send_chunk(sdnow, i);
+					/* send data block */
+					if (!c[sdnow].x_chk[i])
+						send_file(sdnow, i);
 				}
 				/* all partial files sent */
 				if (c[sdnow].pfdo == 0) {
@@ -932,7 +1012,7 @@ main(int argc, char *argv[])
 				    sizeof(in) - 1)) < 1)
 					goto quit;
 			} else {
-				if ((rt = recv(sdnow, in, sizeof(in) - 1, 0))
+				if ((rt = sys_read(sdnow, in, sizeof(in) - 1))
 				    < 1)
 					goto quit;
 			}
@@ -947,16 +1027,21 @@ main(int argc, char *argv[])
 			while (http_header_comp(in) == 0) {
 				tv.tv_sec = 1;
 				tv.tv_usec = 0;
-				if (select(sdnow + 1, &seto_r, NULL, NULL, &tv)
-				    != 1)
+				r = select(sdnow + 1, &seto_r, NULL, NULL, &tv);
+				if (r == -1) {
+					if (errno == EINTR)
+						continue;
+					goto quit;
+				}
+				if (r == 0)
 					goto quit;
 				if (c[sdnow].x_ssl) {
 					if ((rp = SSL_read(c[sdnow].ssl, tmp,
 					    sizeof(tmp) - 1)) < 1)
 						goto quit;
 				} else {
-					if ((rp = recv(sdnow, tmp,
-					    sizeof(tmp) - 1, 0)) < 1)
+					if ((rp = sys_read(sdnow, tmp,
+					    sizeof(tmp) - 1)) < 1)
 						goto quit;
 				}
 				rt += rp;
@@ -969,7 +1054,7 @@ main(int argc, char *argv[])
 						SSL_write(c[sdnow].ssl, b,
 						    strlen(b));
 					} else {
-						send(sdnow, b, strlen(b), 0);
+						sys_write(sdnow, b, strlen(b));
 					}
 					free(h);
 					free(b);
@@ -986,7 +1071,7 @@ main(int argc, char *argv[])
 				if (c[sdnow].x_ssl) {
 					SSL_write(c[sdnow].ssl, b, strlen(b));
 				} else {
-					send(sdnow, b, strlen(b), 0);
+					sys_write(sdnow, b, strlen(b));
 				}
 				free(h);
 				free(b);
@@ -1010,7 +1095,7 @@ main(int argc, char *argv[])
 						SSL_write(c[sdnow].ssl, b,
 						    strlen(b));
 					} else {
-						send(sdnow, b, strlen(b), 0);
+						sys_write(sdnow, b, strlen(b));
 					}
 					free(h);
 					free(b);
@@ -1044,16 +1129,22 @@ main(int argc, char *argv[])
 				while(http_body_comp(header[0], body) != 1) {
 					tv.tv_sec = 1;
 					tv.tv_usec = 0;
-					if (select(sdnow + 1, &seto_r, NULL,
-					    NULL, &tv) != 1)
+					r = select(sdnow + 1, &seto_r, NULL,
+					    NULL, &tv);
+					if (r == -1) {
+						if (errno == EINTR)
+							continue;
+						goto quit;
+					}
+					if (r == 0)
 						goto quit;
 					if (c[sdnow].x_ssl) {
 						if ((rp = SSL_read(c[sdnow].ssl,
 						    tmp, sizeof(tmp) - 1)) < 1)
 							goto quit;
 					} else {
-						if ((rp = recv(sdnow, tmp,
-						    sizeof(tmp) - 1, 0)) < 1)
+						if ((rp = sys_read(sdnow, tmp,
+						    sizeof(tmp) - 1)) < 1)
 							goto quit;
 					}
 					rt += rp;
@@ -1066,8 +1157,8 @@ main(int argc, char *argv[])
 							SSL_write(c[sdnow].ssl,
 							    b, strlen(b));
 						} else {
-							send(sdnow, b,
-							    strlen(b), 0);
+							sys_write(sdnow, b,
+							    strlen(b));
 						}
 						free(h);
 						free(b);
@@ -1094,7 +1185,7 @@ main(int argc, char *argv[])
 						SSL_write(c[sdnow].ssl, b,
 						    strlen(b));
 					} else {
-						send(sdnow, b, strlen(b), 0);
+						sys_write(sdnow, b, strlen(b));
 					}
 					free(h);
 					free(b);
@@ -1109,11 +1200,11 @@ main(int argc, char *argv[])
 			/* check keep-alive */
 			if (r != 1) {
 				quit:
+				if (c[sdnow].x_ssl)
+					SSL_free(c[sdnow].ssl);
 				c[sdnow].to = 0;
 				c[sdnow].x_ip6 = 0;
 				c[sdnow].x_ssl = 0;
-				if (c[sdnow].x_ssl)
-					SSL_free(c[sdnow].ssl);
 				close(sdnow);
 				FD_CLR(sdnow, &master_r);
 			}
