@@ -19,19 +19,21 @@
  */
 void		usage(const int);
 void		sig_handler(const int);
+void		send_chunk(const int, const int);
+void		send_block(const int, const int);
 
 /*
  * http.c
  */
 int		http_verify(const char *, const char *, const int);
-int		http_proc(const char *, char *, int, int);
+int		http_proc(const char *, char *, const int, const int);
 int		http_cgi_getexec(char *, char *, const char *, const int,
 		    const int);
 int		http_cgi_header(char *, char *, const int, const int);
 int		http_header_comp(const char *);
 int		http_body_comp(const char *, char *);
 int		http_access_htaccess(char *, const char *, const int);
-int		http_alog(int, int);
+int		http_alog(const int, const int);
 char *		http_chunk(const char *);
 char *		http_date(struct tm *);
 char *		http_uridecode(const char *);
@@ -45,4 +47,7 @@ struct header *	http_header(const char *, const char *, const int, const int);
 int		sys_mime(char *, const int, const char *, const int,
 		    const char *);
 int		sys_access_auth(const char *, const char *);
+int		sys_read(const int, char *, const int);
+int		sys_write(const int, const char *, const int);
+int		sys_close_except(const int);
 char *		sys_date(struct tm *, int);
